@@ -43,7 +43,7 @@ if typing.TYPE_CHECKING:
 __all__ = ("ScrollBar", "Scrollable", "ScrollableError", "ScrollbarSymbols")
 
 
-WrappedWidget = typing.TypeVar("WrappedWidget")
+WrappedWidget = typing.TypeVar("WrappedWidget", bound="SupportsScroll")
 
 
 class ScrollableError(WidgetError):
@@ -85,10 +85,10 @@ class ScrollbarSymbols(str, enum.Enum):
 
 
 @runtime_checkable
-class SupportsScroll(Protocol):
-    """Protocol for scroll supporting widget.
+class WidgetProto(Protocol):
+    """Protocol for widget.
 
-    Due to protocol can not inherit non-protocol bases, require also several obligatory Widget methods.
+    Due to protocol cannot inherit non-protocol bases, define several obligatory Widget methods.
     """
 
     # Base widget methods (from Widget)
@@ -116,12 +116,38 @@ class SupportsScroll(Protocol):
 
     def render(self, size: tuple[int, int], focus: bool = False) -> Canvas: ...
 
-    # Scroll specific methods
+
+@runtime_checkable
+class SupportsScroll(WidgetProto, Protocol):
+    """Scroll specific methods."""
+
     def get_scrollpos(self, size: tuple[int, int], focus: bool = False) -> int: ...
 
     def rows_max(self, size: tuple[int, int] | None = None, focus: bool = False) -> int: ...
 
 
+@runtime_checkable
+class SupportsRelativeScroll(WidgetProto, Protocol):
+    """Relative scroll-specific methods."""
+
+    def require_relative_scroll(self, size: tuple[int, int], focus: bool = False) -> bool: ...
+
+    def get_first_visible_pos(self, size: tuple[int, int], focus: bool = False) -> int: ...
+
+    def get_visible_amount(self, size: tuple[int, int], focus: bool = False) -> int: ...
+
+
+def orig_iter(w: Widget) -> Iterator[Widget]:
+    visited = {w}
+    yield w
+    while hasattr(w, "original_widget"):
+        w = w.original_widget
+        if w in visited:
+            break
+        visited.add(w)
+        yield w
+
+
 class Scrollable(WidgetDecoration[WrappedWidget]):
     def sizing(self) -> frozenset[Sizing]:
         return frozenset((Sizing.BOX,))
@@ -153,7 +179,11 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
         self.force_forward_keypress = force_forward_keypress
         super().__init__(widget)
 
-    def render(self, size: tuple[int, int], focus: bool = False) -> CompositeCanvas:
+    def render(
+        self,
+        size: tuple[int, int],  # type: ignore[override]
+        focus: bool = False,
+    ) -> CompositeCanvas:
         from urwid import canvas
 
         maxcol, maxrow = size
@@ -236,7 +266,8 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
 
         # Disable cursor display if cursor is outside of visible canvas parts
         if canv.cursor is not None:
-            _curscol, cursrow = canv.cursor  # pylint: disable=unpacking-non-sequence
+            # Pylint check acts here a bit weird.
+            _curscol, cursrow = canv.cursor  # pylint: disable=unpacking-non-sequence,useless-suppression
             if cursrow >= maxrow or cursrow < 0:
                 canv.cursor = None
 
@@ -269,7 +300,11 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
 
         return canv
 
-    def keypress(self, size: tuple[int, int], key: str) -> str | None:
+    def keypress(
+        self,
+        size: tuple[int, int],  # type: ignore[override]
+        key: str,
+    ) -> str | None:
         from urwid.command_map import Command
 
         # Maybe offer key to original widget
@@ -277,7 +312,7 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
             ow = self._original_widget
             ow_size = self._get_original_widget_size(size)
 
-            # Remember previous cursor position if possible
+            # Remember the previous cursor position if possible
             if hasattr(ow, "get_cursor_coords"):
                 self._old_cursor_coords = ow.get_cursor_coords(ow_size)
 
@@ -285,7 +320,7 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
             if key is None:
                 return None
 
-        # Handle up/down, page up/down, etc
+        # Handle up/down, page up/down, etc.
         command_map = self._command_map
         if command_map[key] == Command.UP:
             self._scroll_action = SCROLL_LINE_UP
@@ -310,7 +345,7 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
 
     def mouse_event(
         self,
-        size: tuple[int, int],
+        size: tuple[int, int],  # type: ignore[override]
         event: str,
         button: int,
         col: int,
@@ -374,7 +409,10 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
             elif cursrow >= self._trim_top + maxrow:
                 self._trim_top = max(0, cursrow - maxrow + 1)
 
-    def _get_original_widget_size(self, size: tuple[int, int]) -> tuple[int] | tuple[()]:
+    def _get_original_widget_size(
+        self,
+        size: tuple[int, int],  # type: ignore[override]
+    ) -> tuple[int] | tuple[()]:
         ow = self._original_widget
         sizing = ow.sizing()
         if Sizing.FLOW in sizing:
@@ -426,10 +464,10 @@ class Scrollable(WidgetDecoration[WrappedWidget]):
 class ScrollBar(WidgetDecoration[WrappedWidget]):
     Symbols = ScrollbarSymbols
 
-    def sizing(self):
+    def sizing(self) -> frozenset[Sizing]:
         return frozenset((Sizing.BOX,))
 
-    def selectable(self):
+    def selectable(self) -> bool:
         return True
 
     def __init__(
@@ -454,7 +492,8 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
         """
         if Sizing.BOX not in widget.sizing():
             raise ValueError(f"Not a box widget: {widget!r}")
-        if not isinstance(widget, SupportsScroll):
+
+        if not any(isinstance(w, SupportsScroll) for w in orig_iter(widget)):
             raise TypeError(f"Not a scrollable widget: {widget!r}")
 
         super().__init__(widget)
@@ -464,37 +503,71 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
         self.scrollbar_width = max(1, width)
         self._original_widget_size = (0, 0)
 
-    def render(self, size: tuple[int, int], focus: bool = False) -> CompositeCanvas:
+    def render(
+        self,
+        size: tuple[int, int],  # type: ignore[override]
+        focus: bool = False,
+    ) -> Canvas:
         from urwid import canvas
 
+        def render_no_scrollbar() -> Canvas:
+            self._original_widget_size = size
+            return ow.render(size, focus)
+
+        def render_for_scrollbar() -> Canvas:
+            self._original_widget_size = ow_size
+            return ow.render(ow_size, focus)
+
         maxcol, maxrow = size
 
-        sb_width = self._scrollbar_width
-        ow_size = (max(0, maxcol - sb_width), maxrow)
+        ow_size = (max(0, maxcol - self._scrollbar_width), maxrow)
         sb_width = maxcol - ow_size[0]
 
         ow = self._original_widget
         ow_base = self.scrolling_base_widget
-        ow_rows_max = ow_base.rows_max(size, focus)
-        if ow_rows_max <= maxrow:
-            # Canvas fits without scrolling - no scrollbar needed
-            self._original_widget_size = size
-            return ow.render(size, focus)
-        ow_rows_max = ow_base.rows_max(ow_size, focus)
 
-        ow_canv = ow.render(ow_size, focus)
-        self._original_widget_size = ow_size
+        # Use hasattr instead of protocol: hasattr will return False in case of getattr raise AttributeError
+        # Use __length_hint__ first since it's less resource intensive
+        use_relative = (
+            isinstance(ow_base, SupportsRelativeScroll)
+            and any(hasattr(ow_base, attrib) for attrib in ("__length_hint__", "__len__"))
+            and ow_base.require_relative_scroll(size, focus)
+        )
 
-        pos = ow_base.get_scrollpos(ow_size, focus)
-        posmax = ow_rows_max - maxrow
+        if use_relative:
+            # `operator.length_hint` is Protocol (Spec) over class based and can end false-negative on the instance
+            # use length_hint-like approach with safe `AttributeError` handling
+            ow_len = getattr(ow_base, "__len__", getattr(ow_base, "__length_hint__", lambda: 0))()
+            ow_canv = render_for_scrollbar()
+            visible_amount = ow_base.get_visible_amount(ow_size, focus)
+            pos = ow_base.get_first_visible_pos(ow_size, focus)
 
-        # Thumb shrinks/grows according to the ratio of
-        # <number of visible lines> / <number of total lines>
-        thumb_weight = min(1.0, maxrow / max(1, ow_rows_max))
-        thumb_height = max(1, round(thumb_weight * maxrow))
+            # in the case of estimated length, it can be smaller than real widget length
+            ow_len = max(ow_len, visible_amount, pos)
+            posmax = ow_len - visible_amount
+            thumb_weight = min(1.0, visible_amount / max(1, ow_len))
+
+            if ow_len == visible_amount:
+                # Corner case: formally all contents indexes should be visible, but this does not mean all rows
+                use_relative = False
+
+        if not use_relative:
+            ow_rows_max = ow_base.rows_max(size, focus)
+            if ow_rows_max <= maxrow:
+                # Canvas fits without scrolling - no scrollbar needed
+                return render_no_scrollbar()
+
+            ow_canv = render_for_scrollbar()
+            ow_rows_max = ow_base.rows_max(ow_size, focus)
+            pos = ow_base.get_scrollpos(ow_size, focus)
+            posmax = ow_rows_max - maxrow
+            thumb_weight = min(1.0, maxrow / max(1, ow_rows_max))
+
+        # Thumb shrinks/grows according to the ratio of <number of visible lines> / <number of total lines>
+        thumb_height = max(1, round(thumb_weight * maxrow))  # pylint: disable=possibly-used-before-assignment
 
         # Thumb may only touch top/bottom if the first/last row is visible
-        top_weight = float(pos) / max(1, posmax)
+        top_weight = float(pos) / max(1, posmax)  # pylint: disable=possibly-used-before-assignment
         top_height = int((maxrow - thumb_height) * top_weight)
         if top_height == 0 and top_weight > 0:
             top_height = 1
@@ -518,7 +591,10 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
             ),
         )
 
-        combinelist = [(ow_canv, None, True, ow_size[0]), (sb_canv, None, False, sb_width)]
+        combinelist = [
+            (ow_canv, None, True, ow_size[0]),  # pylint: disable=possibly-used-before-assignment
+            (sb_canv, None, False, sb_width),
+        ]
 
         if self._scrollbar_side != SCROLLBAR_LEFT:
             return canvas.CanvasJoin(combinelist)
@@ -548,15 +624,9 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
         self._invalidate()
 
     @property
-    def scrolling_base_widget(self) -> SupportsScroll:
+    def scrolling_base_widget(self) -> SupportsScroll | SupportsRelativeScroll:
         """Nearest `original_widget` that is compatible with the scrolling API"""
 
-        def orig_iter(w: Widget) -> Iterator[Widget]:
-            while hasattr(w, "original_widget"):
-                w = w.original_widget
-                yield w
-            yield w
-
         w = self
 
         for w in orig_iter(self):
@@ -565,12 +635,16 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
 
         raise ScrollableError(f"Not compatible to be wrapped by ScrollBar: {w!r}")
 
-    def keypress(self, size: tuple[int, int], key: str) -> str | None:
+    def keypress(
+        self,
+        size: tuple[int, int],  # type: ignore[override]
+        key: str,
+    ) -> str | None:
         return self._original_widget.keypress(self._original_widget_size, key)
 
     def mouse_event(
         self,
-        size: tuple[int, int],
+        size: tuple[int, int],  # type: ignore[override]
         event: str,
         button: int,
         col: int,
@@ -579,7 +653,7 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
     ) -> bool | None:
         ow = self._original_widget
         ow_size = self._original_widget_size
-        handled = False
+        handled: bool | None = False
         if hasattr(ow, "mouse_event"):
             handled = ow.mouse_event(ow_size, event, button, col, row, focus)
 
@@ -594,4 +668,4 @@ class ScrollBar(WidgetDecoration[WrappedWidget]):
                 ow.set_scrollpos(pos + 1)
                 return True
 
-        return False
+        return handled
